import typing
import traceback
from .. import SignalSender


class ValueInterface:
    def get_value(self) -> typing.Any:
        pass  # pragma: no cover

    def set_value(self, value) -> None:
        pass  # pragma: no cover


class AttachMixin:
    AttachFlag = '_attach_{}_'

    def attach(self, creator, args: tuple = (), after_created: typing.Callable[[], None] = None):
        name = AttachMixin.AttachFlag.format(self.current_property_name)

        obj = getattr(self, name, None)
        if obj is None:
            obj = creator(*args)
            setattr(self, name, obj)
            if after_created:
                after_created()
        return obj

    def assign(self, value):
        name = AttachMixin.AttachFlag.format(self.current_property_name)
        setattr(self, name, value)

    @property
    def current_property_name(self):
        return traceback.extract_stack(None, 3)[0][2]


class ChangedMixin(ValueInterface, AttachMixin):
    @property
    def changed(self) -> SignalSender:
        return self.attach(creator=SignalSender, after_created=self.set_changed_connection)

    def check_change(self, *_):
        self.changed.emit_if_changed(self.get_value())

    def set_changed_connection(self):
        pass  # pragma: no cover


class ValueEntity(ValueInterface):
    pass


class ChangedValueEntity(ValueEntity, ChangedMixin):
    def connect(self, other: 'ChangedValueEntity'):
        self.changed.connect(other.set_value)
        other.changed.connect(self.set_value)
        other.set_value(self.get_value())


class AbstractProperty(ChangedValueEntity):
    def __init__(self,
                 source_getter: typing.Callable[[], typing.Any],
                 source_setter: typing.Callable[[typing.Any], None] = None,
                 source_changed: SignalSender = None
                 ):
        self.source_getter = source_getter
        self.source_setter = source_setter
        self.source_changed = source_changed

    def set_changed_connection(self) -> None:
        if self.source_changed is not None:
            self.source_changed.connect(self.check_change)

    def get_value(self) -> typing.Any:
        return self.source_getter()

    def set_value(self, value) -> None:
        self.source_setter(value)


class AbstractContainer(ChangedValueEntity):
    def get_property(self, name) -> typing.Any:
        raise NotImplementedError

    def set_property(self, name, value) -> None:
        raise NotImplementedError

    # children property
    @property
    def children(self) -> typing.List['AbstractItem']:
        return self.attach(list)

    def add(self, item: 'AbstractItem') -> None:
        self.children.append(item)
        self.changed.connect(item.check_change)

    # name property
    def __setattr__(self, key: str, value) -> None:
        if isinstance(value, AbstractItem):
            value.name = key
        super().__setattr__(key, value)


class AbstractItem(ChangedValueEntity):
    def __init__(self, container: AbstractContainer = None):
        self.name: str = None
        self._storage: typing.Any = None

        self.container = container
        if container is not None:
            container.add(self)

    def get_value(self) -> typing.Any:
        if self.container is not None:
            return self.container.get_property(self.name)
        else:
            return self._storage

    def set_value(self, value) -> None:
        if self.container is not None:
            self.container.set_property(self.name, value)
        else:
            self._storage = value
            self.check_change()



import os
import sys
import PySide2
from PySide2 import QtCore, QtWidgets

pyside2_plugins = os.path.join(os.path.dirname(PySide2.__file__), "plugins")
QtWidgets.QApplication.addLibraryPath(pyside2_plugins)

Qt = QtCore.Qt
# noinspection PyArgumentList
qt_app = QtWidgets.QApplication.instance()

if not qt_app:
    if getattr(sys, 'frozen', None):  # pragma: no cover
        # noinspection PyArgumentList
        QtWidgets.QApplication.addLibraryPath(QtCore.QDir.currentPath())

    qt_app = QtWidgets.QApplication([])



from . import BaseWidget
from ...model import WidgetStringItem
from PySide2.QtWidgets import QSpinBox


class SpinBox(QSpinBox, BaseWidget, WidgetStringItem):
    def get_string_value(self) -> str:
        return str(self.get_integer_value())

    def set_string_value(self, value: str):
        self.set_integer_value(int(value))

    def get_integer_value(self) -> int:
        return self.value()

    def set_integer_value(self, value: int):
        self.setValue(value)

    def set_changed_connection(self):
        # noinspection PyUnresolvedReferences
        self.valueChanged[int].connect(self.integer.check_change)
        # noinspection PyUnresolvedReferences
        self.valueChanged[str].connect(self.string.check_change)



from .base import AbstractItem, ChangedValueEntity
from .types import StringItem, IntegerItem, FloatItem
from .types import IntegerTypeProperty


class WidgetStringItem(StringItem, IntegerItem, FloatItem):
    def get_integer_value(self) -> int:
        try:
            return int(self.get_string_value())
        except ValueError:
            return 0

    def set_integer_value(self, value: int):
        self.set_string_value(str(value))

    def get_float_value(self) -> float:
        try:
            return float(self.get_string_value())
        except ValueError:
            return 0

    def set_float_value(self, value: float):
        self.set_string_value(str(value))


class IndexPropertyInterface(ChangedValueEntity):
    @property
    def index(self) -> IntegerTypeProperty:
        return self.attach(creator=IntegerTypeProperty, args=(
            self.get_index_value, self.set_index_value, self.changed
        ))

    def get_index_value(self) -> int:
        raise NotImplementedError

    def set_index_value(self, value: int):
        raise NotImplementedError


class WidgetIndexItem(AbstractItem, IndexPropertyInterface):
    def get_index_value(self) -> int:
        raise NotImplementedError

    def set_index_value(self, value: int):
        raise NotImplementedError



from . import BaseWidget
from ...model import WidgetStringItem
from PySide2.QtWidgets import QLineEdit


class LineEdit(QLineEdit, BaseWidget, WidgetStringItem):
    def get_value(self) -> str:
        return self.text()

    def set_value(self, value: str):
        value = value or ''
        if value != self.get_value():
            self.setText(value)

    def set_changed_connection(self):
        # noinspection PyUnresolvedReferences
        self.textChanged.connect(self.check_change)



import typing
from ..core import Qt
from .utils import scaling_ratio
from PySide2.QtGui import QPen, QColor, QBrush


class Pen(QPen):
    def __init__(self,
                 para: typing.Union[QBrush, QColor, QPen],
                 width: typing.Optional[float] = None,
                 pen_style: Qt.PenStyle = Qt.SolidLine,
                 cap_style: Qt.PenCapStyle = Qt.SquareCap,
                 join_style: Qt.PenJoinStyle = Qt.BevelJoin
                 ):
        if width is None:
            super().__init__(para)
            self.width *= scaling_ratio
        else:
            super().__init__(para, width * scaling_ratio, pen_style, cap_style, join_style)

    @property
    def width(self) -> float:
        return self.widthF() / scaling_ratio

    @width.setter
    def width(self, value: float):
        self.setWidthF(value * scaling_ratio)



import typing
from . import BaseWidget
from PySide2.QtGui import QKeySequence
from PySide2.QtWidgets import QShortcut
from ..interfaces import ExcitedSignalInterface


class Shortcut(QShortcut, BaseWidget, ExcitedSignalInterface):  # pragma: no cover
    def __init__(self, key: typing.Union[str, QKeySequence], parent):
        if isinstance(key, str):
            key = QKeySequence(key)
        super().__init__(key, parent)

    def set_excited_signal_connection(self):
        # noinspection PyUnresolvedReferences
        self.activated.connect(self.excited.emit)



from setuptools import setup, find_packages


setup(
    name='quina',
    version='0.0.7',
    description='PySide2 MVVM Framework',
    url='https://github.com/KD-Group/quina',
    author='KD-Group',
    author_email='sfzhou.scut@gmail.com',

    license='GPL-3.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
    ],

    keywords='PySide2 MVVM',
    packages=find_packages(exclude=['docs', 'tests', 'examples']),
    install_requires=['PySide2', 'typing'],
    extras_require={'test': ['flake8', 'pytest', 'pytest-cov']}
)
